#!/usr/local/bin/python
# encoding: utf-8
"""
*Continually listen to gracedb and await for new waves and new wave maps*

:Author:
    David Young

:Date Created:
    June 22, 2016
"""
################# GLOBAL IMPORTS ####################
import sys
import os
os.environ['TERM'] = 'vt100'
from fundamentals import tools
import numpy as np
from ligo.gracedb.rest import GraceDb, HTTPError
from ligo.gracedb.rest import GraceDbBasic
from astrocalc.times import conversions
from astrocalc.times import now as mjdNow
import time


class listen():
    """
    *The worker class for the listen module*

    **Key Arguments:**
        - ``log`` -- logger
        - ``settings`` -- the settings dictionary
        - ``label`` -- filter wave event by label. Default *EM_READY*
        - ``farThreshold`` -- the false alarm rate threshold. Default *1e-7*
        - ``startMJD`` -- startMJD. Default *56658.0 (2014-01-01)*
        - ``endMJD`` -- endMJD. Default *69807.0 (2050-01-01)*
        - ``daemon`` -- run in daemon mode. Ignores MJD settings and pings GraceDB every 1 mins.
    """
    # Initialisation

    def __init__(
            self,
            log,
            settings=False,
            label="EM_READY",
            farThreshold=1e-7,
            startMJD=56658.0,
            endMJD=69807.0,
            daemon=False
    ):
        self.log = log
        log.debug("instansiating a new 'listen' object")
        self.settings = settings
        self.label = label
        self.farThreshold = farThreshold
        self.startMJD = startMJD
        self.endMJD = endMJD
        self.daemon = daemon
        # xt-self-arg-tmpx

        # Initial Actions
        self.mapDirectory = self.settings["gw maps directory"]
        # RECURSIVELY CREATE MISSING DIRECTORIES
        if not os.path.exists(self.mapDirectory):
            os.makedirs(self.mapDirectory)

        # INSTANTIATE GRACEDB CLIENT WITH A QUERY STRING
        self.client = GraceDbBasic()

        return None

    def get_maps(
            self):
        """*download the maps for all events in our time range*

        **Usage:**

            .. code-block:: python

                from breaker.gracedb import listen
                downloader = listen(
                    log=log,
                    settings=settings,
                    label="EM_READY",
                    farThreshold=1e-7,
                    startMJD=56658.0,
                    endMJD=69807.0
                )
                downloader.get_maps()
        """
        self.log.info('starting the ``get_maps`` method')

        eventKeys = ['graceid', 'gpstime', 'group', 'links', 'created',
                     'far', 'instruments', 'labels', 'nevents', 'submitter', 'search', 'likelihood']
        fileorder = ['LALInference_skymap.fits.gz',
                     'bayestar.fits.gz', 'LIB_skymap.fits.gz', 'skymap.fits.gz']

        # CONVERT MJD TO GPS
        converter = conversions(
            log=self.log
        )
        gpsZeroMjd = converter.ut_datetime_to_mjd(
            utDatetime="1980-01-06 00:00:00")

        stop = False

        beginningOfTime = 1e09
        while stop == False:

            if self.daemon:
                mjd = mjdNow(
                    log=self.log
                ).get_mjd()
                # 20 MINS AGO
                startGPS = (mjd - float(gpsZeroMjd)) * 60 * 60 * 24. - \
                    beginningOfTime
                beginningOfTime = 0.
                # 20 MINS FROM NOW
                endGPS = (mjd - float(gpsZeroMjd)) * 60 * 60 * 24. + 1200.
            else:
                stop = True
                startGPS = (self.startMJD - float(gpsZeroMjd)) * 60 * 60 * 24.
                endGPS = (self.endMJD - float(gpsZeroMjd)) * 60 * 60 * 24.

            label = self.label
            farThreshold = self.farThreshold
            self.eventString = '%(label)s far <%(farThreshold)s %(startGPS)s  .. %(endGPS)s' % locals(
            )

            # REST API RETURNS AN ITERATOR
            self.events = self.client.events(
                query=self.eventString,
                orderby=None,
                count=None,
                columns=None
            )

            count = 0
            for event in self.events:
                count += 1
                eventinfo = {}
                for key in eventKeys:
                    if not key in event:
                        self.log.warning(
                            "`%(key)s` not in event %(event)s" % locals())
                        continue
                    eventinfo[key] = event[key]
                eventinfo['gpstime'] = float(eventinfo['gpstime'])

                if eventinfo['far'] > self.farThreshold:
                    far = eventinfo['far']
                    farthres = self.farThreshold
                    self.info.warning(
                        "event %(event)s does not pass FAR threshold of %(farthres)s. (FAR = %(far)s)" % locals())
                    continue

                allMaps = []

                for lvfile in fileorder:
                    try:
                        aMap = self.client.files(eventinfo['graceid'], lvfile)
                        allMaps.append(aMap)
                        self._write_map_to_disk(
                            sMap=aMap,
                            mapName=lvfile,
                            waveId=eventinfo['graceid']
                        )
                    except:
                        eventId = eventinfo['graceid']
                        self.log.info(
                            "The %(lvfile)s path for %(eventId)s does not seem to exist yet" % locals())

                if len(allMaps) == 0:
                    eventId = eventinfo['graceid']
                    self.log.warning(
                        'cound not download skymaps for event %(eventId)s' % locals())

            if stop == False:
                if count == 0:
                    print "No recent events, will try again in 60 secs"
                else:
                    print "%(count)s recent events found, will try again in 60 secs" % locals()
                # WAIT 1 MIN
                time.sleep(60)

        self.log.info('completed the ``get_maps`` method')
        return None

    def _write_map_to_disk(
            self,
            sMap,
            mapName,
            waveId):
        """*Given a skymap, write it to disk*

        **Key Arguments:**
            - ``sMap`` -- the skymap from gracedb.
            - ``mapName`` -- the map flavour.
            - ``waveId`` -- the graceDB id for the wave.
        """
        self.log.info('starting the ``_write_map_to_disk`` method')

        outputDir = self.mapDirectory + "/" + waveId.upper()
        # RECURSIVELY CREATE MISSING DIRECTORIES
        if not os.path.exists(outputDir):
            os.makedirs(outputDir)

        print "Downloading %(mapName)s for GW event %(waveId)s " % locals()

        mapPath = outputDir + "/" + mapName

        skymapfile = open(mapPath, 'w')
        skymapfile.write(sMap.read())
        skymapfile.close()

        self.log.info('completed the ``_write_map_to_disk`` method')
        return None



#!/usr/local/bin/python
# encoding: utf-8
"""
*Class to setup database object for the breaker*

:Author:
    David Young

:Date Created:
    October 29, 2015
"""
################# GLOBAL IMPORTS ####################
import sys
import os
os.environ['TERM'] = 'vt100'
import readline
import glob
import pickle
import time
import MySQLdb as ms
from docopt import docopt
from fundamentals import tools, times


class database():

    """
    *The worker class for the database module*

    **Key Arguments:**
        - ``log`` -- logger
        - ``settings`` -- the settings dictionary
    """
    # INITIALISATION

    def __init__(
            self,
            log,
            settings=False,

    ):
        self.log = log
        log.debug("instansiating a new '_database' object")
        self.settings = settings
        return None

    def close(self):
        del self
        return None

    # METHOD ATTRIBUTES
    def get(self):
        """
        *get the database object*

        **Return:**
            - ``self.transientsDbConn, self.ps1gwDbConn, self.cataloguesDbConn`` -- three database connections
        """
        self.log.debug('starting the ``get`` method')
        self._setup_database_connections()
        self.log.debug('completed the ``get`` method')
        return self.ligo_virgo_wavesDbConn, self.ps1gwDbConn, self.cataloguesDbConn

    def _setup_database_connections(
            self):
        """
        *setup database connections for transient and catalogue databases*
        """
        self.log.debug('starting the ``_setup_database_connections`` method')

        from subprocess import Popen, PIPE, STDOUT

        tunnelDatabases = {}
        for db in self.settings["database settings"]:
            port = self.settings["database settings"][db]["port"]
            if "tunnel" in str(port):
                if port not in tunnelDatabases.keys():
                    tunnelDatabases[port] = []
                tunnelDatabases[port].append(
                    db)

        # SETUP TUNNEL IF REQUIRED
        if "ssh tunnels" in self.settings:
            tunnels = self.settings["ssh tunnels"]
            for tunnelName in tunnels:
                tunnel = self.settings["ssh tunnels"][tunnelName]
                if tunnel["use tunnel"] is True:
                    # TEST TUNNEL DOES NOT ALREADY EXIST
                    sshPort = tunnel["port"]
                    for db in tunnelDatabases[tunnelName]:
                        connected = self._checkServer(
                            self.settings["database settings"][db]["host"], sshPort)
                        if connected:
                            break
                    if connected:
                        self.log.debug('ssh tunnel already exists - moving on')
                    else:
                        # GRAB TUNNEL SETTINGS FROM SETTINGS FILE
                        ru = tunnel["remote user"]
                        rip = tunnel["remote ip"]
                        rh = tunnel["remote datbase host"]

                        cmd = "ssh -fnN %(ru)s@%(rip)s -L %(sshPort)s:%(rh)s:3306" % locals()
                        p = Popen(cmd, shell=True, close_fds=True)
                        output = p.communicate()[0]
                        self.log.debug('output: %(output)s' % locals())

                        # TEST CONNECTION - QUIT AFTER SO MANY TRIES
                        connected = False
                        count = 0
                        while not connected:
                            for db in tunnelDatabases[tunnelName]:
                                connected = self._checkServer(
                                    self.settings["database settings"][db]["host"], sshPort)
                                if connected:
                                    break
                            time.sleep(1)
                            count += 1
                            if count == 5:
                                self.log.error(
                                    'cound not setup tunnel to remote datbase' % locals())
                                sys.exit(0)

        # SETUP A DATABASE CONNECTION FOR THE ps1gw
        host = self.settings["database settings"][
            "ps1gw"]["host"]
        user = self.settings["database settings"][
            "ps1gw"]["user"]
        passwd = self.settings["database settings"][
            "ps1gw"]["password"]
        dbName = self.settings["database settings"][
            "ps1gw"]["db"]
        port = self.settings["database settings"][
            "ps1gw"]["port"]
        if "tunnel" in str(port):
            port = self.settings["ssh tunnels"][port]["port"]
        thisConn = ms.connect(
            host=host,
            user=user,
            passwd=passwd,
            db=dbName,
            port=port,
            use_unicode=True,
            charset='utf8'
        )
        thisConn.autocommit(True)
        self.log.debug('ps1gwDbConn: %s' % (thisConn,))
        self.ps1gwDbConn = thisConn

        # SETUP DATABASE CONNECTION FOR WAVE DATABASE
        host = self.settings["database settings"][
            "ligo_virgo_waves"]["host"]
        user = self.settings["database settings"][
            "ligo_virgo_waves"]["user"]
        passwd = self.settings["database settings"][
            "ligo_virgo_waves"]["password"]
        dbName = self.settings["database settings"][
            "ligo_virgo_waves"]["db"]
        port = self.settings["database settings"][
            "ligo_virgo_waves"]["port"]
        if "tunnel" in str(port):
            port = self.settings["ssh tunnels"][port]["port"]
        thisConn = ms.connect(
            host=host,
            user=user,
            passwd=passwd,
            db=dbName,
            port=port,
            use_unicode=True,
            charset='utf8'
        )
        thisConn.autocommit(True)
        self.log.debug('ligo_virgo_wavesDbConn: %s' % (thisConn,))
        self.ligo_virgo_wavesDbConn = thisConn

        # SETUP DATABASE CONNECTION FOR CATALOGUE DATABASE
        host = self.settings["database settings"][
            "catalogues"]["host"]
        user = self.settings["database settings"][
            "catalogues"]["user"]
        passwd = self.settings["database settings"][
            "catalogues"]["password"]
        dbName = self.settings["database settings"][
            "catalogues"]["db"]
        port = self.settings["database settings"][
            "catalogues"]["port"]
        if "tunnel" in str(port):
            port = self.settings["ssh tunnels"][port]["port"]
        thisConn = ms.connect(
            host=host,
            user=user,
            passwd=passwd,
            db=dbName,
            port=port,
            use_unicode=True,
            charset='utf8'
        )
        thisConn.autocommit(True)
        self.log.debug('catalogues database connection: %s' % (thisConn,))
        self.cataloguesDbConn = thisConn

        self.log.debug('completed the ``_setup_database_connections`` method')
        return None

    def _checkServer(self, address, port):
        """
        *Check that the TCP Port we've decided to use for tunnelling is available*
        """
        self.log.debug('starting the ``_checkServer`` method')

        # CREATE A TCP SOCKET
        import socket
        s = socket.socket()
        self.log.debug(
            """Attempting to connect to `%(address)s` on port `%(port)s`""" % locals())
        try:
            s.connect((address, port))
            self.log.debug(
                """Connected to `%(address)s` on port `%(port)s`""" % locals())
            return True
        except socket.error, e:
            self.log.warning(
                """Connection to `%(address)s` on port `%(port)s` failed - try again: %(e)s""" % locals())
            return False

        return None

    # xt-class-method


if __name__ == '__main__':
    main()



#!/usr/local/bin/python
# encoding: utf-8
"""
*Generate stats for a gravitational wave survey campaign footprint*
"""
################# GLOBAL IMPORTS ####################
# SUPPRESS MATPLOTLIB WARNINGS
import warnings
warnings.filterwarnings("ignore")
import sys
import os
os.environ['TERM'] = 'vt100'
import healpy as hp
import numpy as np
import math
import matplotlib
import matplotlib.cm as cm
import matplotlib.pyplot as plt
from matplotlib.path import Path
from matplotlib.pyplot import savefig
import matplotlib.patches as patches
from fundamentals import tools, times
from astrocalc.coords import separations


class survey_footprint():
    """
    *Generate stats for a gravitational wave survey campaign footprint*

    **Key Arguments:**
        - ``log`` -- logger
        - ``settings`` -- the settings dictionary
        - ``gwid`` -- the unique wave id

    **Usage:**

        To generate cummulative stats for the wave "G184098"

        .. code-block:: python 

            from breaker.stats import survey_footprint
            stats = survey_footprint(
                log=log,
                settings=settings,
                gwid="G184098"
            )
            stats.get()
    """
    # Initialisation

    def __init__(
            self,
            log,
            settings=False,
            gwid=False

    ):
        self.log = log
        log.debug("instansiating a new 'survey_footprint' object")
        self.settings = settings
        self.gwid = gwid

        # INITIAL ACTIONS - CONNECT TO THE DATABASE REQUIRED
        from breaker import database
        db = database(
            log=self.log,
            settings=self.settings
        )
        self.ligo_virgo_wavesDbConn, self.ps1gwDbConn, self.cataloguesDbConn = db.get()

        return None

    def get(self):
        """
        *get the survey footprint stats and print to screen/file*

        **Return:**
            - ``None``
        """
        self.log.info('starting the ``get`` method')

        # GRAB METADATA FROM THE DATABASES
        from breaker.plots import plot_wave_observational_timelines
        this = plot_wave_observational_timelines(
            log=self.log, settings=self.settings)
        plotParameters, ps1Transients, ps1Pointings, altasPointings = this.get_gw_parameters_from_settings(
            gwid=self.gwid)

        # GET THE PROBABILITY MAP FOR THE GIVEN GWID
        pathToProbMap = self.settings[
            "gravitational waves"][self.gwid]["mapPath"]
        mapName = pathToProbMap.split("/")[-1].replace(".fits", "")
        if not os.path.exists(pathToProbMap):
            message = "the path to the map %s does not exist on this machine" % (
                pathToProbMap,)
            self.log.critical(message)
            raise IOError(message)

        # UNPACK THE PLOT PARAMETERS
        pixelSizeDeg = 0.066667
        centralCoordinate = plotParameters["centralCoordinate"]
        raRange = plotParameters["raRange"]
        decRange = plotParameters["decRange"]

        raMax = centralCoordinate[0] + raRange / 2.
        raMin = centralCoordinate[0] - raRange / 2.
        decMax = centralCoordinate[1] + decRange / 2.
        decMin = centralCoordinate[1] - decRange / 2.

        # DETERMINE THE PIXEL GRID X,Y RANGES
        xRange = int(raRange / pixelSizeDeg)
        yRange = int(decRange / pixelSizeDeg)

        # CREATE A NEW WCS OBJECT
        import numpy
        from astropy import wcs as awcs
        from astropy.io import fits
        w = awcs.WCS(naxis=2)

        # SET THE REFERENCE PIXEL TO THE CENTRE PIXEL
        w.wcs.crpix = [xRange / 2., yRange / 2.]
        # SET THE REQUIRED PIXEL SIZE
        w.wcs.cdelt = numpy.array([pixelSizeDeg, pixelSizeDeg])
        # WORLD COORDINATES AT REFERENCE PIXEL
        w.wcs.crval = centralCoordinate
        # USE THE "GNOMONIC" PROJECTION ("COORDINATESYS---PROJECTION")
        w.wcs.ctype = ["RA---TAN", "DEC--TAN"]

        # CREATE THE FITS HEADER WITH WCS
        header = w.to_header()

        # CREATE A PIXEL GRID - 2 ARRAYS OF X, Y
        columns = []
        px = np.tile(np.arange(0, xRange), yRange)
        py = np.repeat(np.arange(0, yRange), xRange)

        # CONVERT THE PIXELS TO WORLD COORDINATES
        wr, wd = w.wcs_pix2world(px, py, 1)

        # MAKE SURE RA IS +VE
        nr = []
        nr[:] = [r if r > 0 else r + 360. for r in wr]
        wr = np.array(nr)

        # READ HEALPIX MAPS FROM FITS FILE
        # THIS FILE IS A ONE COLUMN FITS BINARY, WITH EACH CELL CONTAINING AN
        # ARRAY OF PROBABILITIES (3,072 ROWS)
        import healpy as hp
        aMap, mapHeader = hp.read_map(pathToProbMap, h=True)
        nside = hp.pixelfunc.get_nside(aMap)
        hpixArea = hp.nside2pixarea(nside, "degrees")

        # import matplotlib.pyplot as plt
        # hp.mollview(aMap, title="mollview image RING", cmap="YlOrRd")
        # hp.graticule()
        # plt.show()

        # HEALPY REQUIRES RA, DEC IN RADIANS AND AS TWO SEPERATE ARRAYS
        import math
        pi = (4 * math.atan(1.0))
        DEG_TO_RAD_FACTOR = pi / 180.0
        RAD_TO_DEG_FACTOR = 180.0 / pi

        # THETA: IS THE POLAR ANGLE, RANGING FROM 0 AT THE NORTH POLE TO PI AT THE SOUTH POLE.
        # PHI: THE AZIMUTHAL ANGLE ON THE SPHERE FROM 0 TO 2PI
        # CONVERT DEC TO THE REQUIRED HEALPIX FORMAT
        nd = -wd + 90.

        # CONVERT WORLD TO HEALPIX INDICES (NON-UNIQUE IDS!)
        healpixIds = hp.ang2pix(nside, theta=nd * DEG_TO_RAD_FACTOR,
                                phi=wr * DEG_TO_RAD_FACTOR)

        # NOW READ THE VALUES OF THE MAP AT THESE HEALPIX INDICES
        probs = []
        probs[:] = [aMap[i] for i in healpixIds]
        probs = np.array(probs)

        # CREATE THE FITS FILE
        hdu = fits.PrimaryHDU(header=header, data=probs)

        # SORT ALL POINTINGS VIA MJD
        from operator import itemgetter
        ps1Pointings = list(ps1Pointings)
        for pt in ps1Pointings:
            pt["survey"] = "ps1"

        altasPointings = list(altasPointings)
        for at in altasPointings:
            at["survey"] = "atlas"

        allPointings = ps1Pointings + altasPointings
        allPointings = sorted(
            allPointings, key=itemgetter('mjd'), reverse=False)

        # ONLY KEEP NON-ZERO PROB FOTTPRINTS
        tmpPointings = []
        for pt in allPointings:
            pra = pt["raDeg"]
            pdec = pt["decDeg"]
            # REMOVE LOWER PROBABILITY FOOTPRINTS
            phi = pra
            if phi > 180.:
                phi = phi - 360.
            theta = -pdec + 90.
            # CONVERT WORLD TO HEALPIX INDICES (NON-UNIQUE IDS!)
            healpixId = hp.ang2pix(nside, theta=theta * DEG_TO_RAD_FACTOR,
                                   phi=phi * DEG_TO_RAD_FACTOR)
            thisProb = aMap[healpixId]
            thisProb = float("%0.*f" % (7, thisProb))
            if thisProb != 0.:
                tmpPointings.append(pt)
        allPointings = tmpPointings

        ps1exposureRadius = 1.4
        atlasPointingSide = 5.46
        healpixDictionary = {}
        totalPointings = len(allPointings)
        cumCoveredArray = np.ones(healpixIds.size)
        finalMJDs = []
        finalAreas = []
        finalProbs = []

        for pti, pt in enumerate(allPointings):

            pra = pt["raDeg"]
            pdec = pt["decDeg"]
            pmjd = pt["mjd"]
            psurvey = pt["survey"]

            decFactor = math.cos(pdec * DEG_TO_RAD_FACTOR)
            coveredPixels = []

            covered = -1
            for r, d, c in zip(wr, wd, cumCoveredArray):
                if psurvey == "ps1":
                    if c == 0:
                        # WE HAVE COVERED THIS PIXEL BEFORE -- MOVE ON
                        covered = 0
                    elif (((pra - r) * decFactor)**2)**0.5 > ps1exposureRadius:
                        covered = 0
                    elif ((pdec - d)**2)**0.5 > ps1exposureRadius:
                        covered = 0
                    else:
                        # CALCULATE SEPARATION IN ARCSEC

                        calculator = separations(
                            log=self.log,
                            ra1=pra,
                            dec1=pdec,
                            ra2=r,
                            dec2=d,
                        )
                        angularSeparation, northSep, eastSep = calculator.get()
                        if angularSeparation < ps1exposureRadius * 60 * 60:
                            covered = 1
                        else:
                            covered = 0

                elif psurvey == "atlas":
                    atDecFactor = math.cos(d * DEG_TO_RAD_FACTOR)
                    if c == 0:
                        # WE HAVE COVERED THIS PIXEL BEFORE -- MOVE ON
                        covered = 0
                    elif ((pdec - d)**2)**0.5 > atlasPointingSide / 2:
                        covered = 0
                    elif (((pra - r) * atDecFactor)**2)**0.5 > atlasPointingSide / 2:
                        covered = 0
                    else:
                        covered = 1
                coveredPixels.append(covered)

            # APPEND TO CUMMULATIVE COVERED ARRAY - SPEED THINGS UP!
            cumCoveredArray = cumCoveredArray * (1 - np.array(coveredPixels))

            # RESHAPE THE ARRAY AS BITMAP
            coveredPixelGrid = np.reshape(
                np.array(coveredPixels), (yRange, xRange))

            # GRAB THE WCS FROM HEADER GENERATED EARLIER
            from wcsaxes import datasets, WCS
            wcs = WCS(hdu.header)

            # PLOT MAP WITH PROJECTION IN HEADER
            import matplotlib.pyplot as plt
            fig = plt.figure()
            ax = fig.add_axes([0.15, 0.1, 0.8, 0.8], projection=wcs)
            im = ax.imshow(coveredPixelGrid,
                           cmap="YlOrRd", origin='lower', alpha=0.7, zorder=1)

            coveredHealPixels = coveredPixels * healpixIds
            coveredProbabilities = coveredPixels * probs

            for cx, cp in zip(coveredHealPixels, coveredProbabilities):
                if cx != 0:
                    healpixDictionary[cx] = cp

            cumArea = len(healpixDictionary) * hpixArea
            cumProb = sum(healpixDictionary.values())

            finalMJDs.append(pmjd)
            finalAreas.append(cumArea)
            finalProbs.append(cumProb)

            print "%(pti)s/%(totalPointings)s.  MJD: %(pmjd)s. AREA: %(cumArea)0.2f. PROB: %(cumProb)0.5f. SURVEY: %(psurvey)s" % locals()

        gwid = self.gwid
        import codecs
        from datetime import datetime, date, time
        now = datetime.now()
        now = now.strftime("%Y%m%dt%H%M%S")
        pathToWriteFile = "/Users/Dave/Desktop/%(now)s_%(gwid)s_%(mapName)s.csv" % locals(
        )
        writeFile = codecs.open(pathToWriteFile, encoding='utf-8', mode='w')
        writeFile.write(
            "MJD,Cummulative Area (degrees),Cummulative Probability Covered\n" % locals())
        for m, a, p in zip(finalMJDs, finalAreas, finalProbs):
            writeFile.write("%(m)0.3f,%(a)0.2f,%(p)0.5f\n" % locals())
        writeFile.close()

        self.log.info('completed the ``get`` method')
        return None

    # xt-class-method
